from dataclasses import dataclass
from typing import TYPE_CHECKING, Optional

from pydantic import Field, SecretStr

from unstructured_ingest.embed.interfaces import (
    AsyncBaseEmbeddingEncoder,
    BaseEmbeddingEncoder,
    EmbeddingConfig,
)
from unstructured_ingest.logger import logger
from unstructured_ingest.utils.dep_check import requires_dependencies
from unstructured_ingest.v2.errors import ProviderError, UserAuthError, UserError, is_internal_error
from unstructured_ingest.v2.errors import (
    RateLimitError as CustomRateLimitError,
)

if TYPE_CHECKING:
    from voyageai import AsyncClient as AsyncVoyageAIClient
    from voyageai import Client as VoyageAIClient


class VoyageAIEmbeddingConfig(EmbeddingConfig):
    batch_size: int = Field(
        default=32,
        le=128,
        description="Batch size for embedding requests. VoyageAI has a limit of 128.",
    )
    api_key: SecretStr
    embedder_model_name: str = Field(default="voyage-3", alias="model_name")
    truncation: Optional[bool] = Field(default=None)
    max_retries: int = 0
    timeout_in_seconds: Optional[int] = None

    def wrap_error(self, e: Exception) -> Exception:
        if is_internal_error(e=e):
            return e
        # https://docs.voyageai.com/docs/error-codes
        from voyageai.error import AuthenticationError, RateLimitError, VoyageError

        if not isinstance(e, VoyageError):
            logger.error(f"unhandled exception from openai: {e}", exc_info=True)
            raise e
        http_code = e.http_status
        message = e.user_message
        if isinstance(e, AuthenticationError):
            return UserAuthError(message)
        if isinstance(e, RateLimitError):
            return CustomRateLimitError(message)
        if 400 <= http_code < 500:
            return UserError(message)
        if http_code >= 500:
            return ProviderError(message)
        logger.error(f"unhandled exception from openai: {e}", exc_info=True)
        return e

    @requires_dependencies(
        ["voyageai"],
        extras="embed-voyageai",
    )
    def get_client(self) -> "VoyageAIClient":
        """Creates a VoyageAI python client to embed elements."""
        from voyageai import Client as VoyageAIClient

        client = VoyageAIClient(
            api_key=self.api_key.get_secret_value(),
            max_retries=self.max_retries,
            timeout=self.timeout_in_seconds,
        )
        return client

    @requires_dependencies(
        ["voyageai"],
        extras="embed-voyageai",
    )
    def get_async_client(self) -> "AsyncVoyageAIClient":
        """Creates a VoyageAI python client to embed elements."""
        from voyageai import AsyncClient as AsyncVoyageAIClient

        client = AsyncVoyageAIClient(
            api_key=self.api_key.get_secret_value(),
            max_retries=self.max_retries,
            timeout=self.timeout_in_seconds,
        )
        return client


@dataclass
class VoyageAIEmbeddingEncoder(BaseEmbeddingEncoder):
    config: VoyageAIEmbeddingConfig

    def wrap_error(self, e: Exception) -> Exception:
        return self.config.wrap_error(e=e)

    def get_client(self) -> "VoyageAIClient":
        return self.config.get_client()

    def embed_batch(self, client: "VoyageAIClient", batch: list[str]) -> list[list[float]]:
        response = client.embed(texts=batch, model=self.config.embedder_model_name)
        return response.embeddings


@dataclass
class AsyncVoyageAIEmbeddingEncoder(AsyncBaseEmbeddingEncoder):
    config: VoyageAIEmbeddingConfig

    def wrap_error(self, e: Exception) -> Exception:
        return self.config.wrap_error(e=e)

    def get_client(self) -> "AsyncVoyageAIClient":
        return self.config.get_async_client()

    async def embed_batch(
        self, client: "AsyncVoyageAIClient", batch: list[str]
    ) -> list[list[float]]:
        response = await client.embed(texts=batch, model=self.config.embedder_model_name)
        return response.embeddings



import json
import typing as t
from dataclasses import dataclass, field
from datetime import datetime

import pytest
import pytz

from unstructured_ingest.cli.utils import extract_config
from unstructured_ingest.interfaces import BaseConfig
from unstructured_ingest.utils.string_and_date_utils import (
    ensure_isoformat_datetime,
    json_to_dict,
    truncate_string_bytes,
)


@dataclass
class A(BaseConfig):
    a: str


@dataclass
class B(BaseConfig):
    a: A
    b: int


flat_data = {"a": "test", "b": 4, "c": True}


def test_extract_config_concrete():
    @dataclass
    class C(BaseConfig):
        b: B
        c: bool

    c = extract_config(flat_data=flat_data, config=C)
    expected_result = {"b": {"a": {"a": "test"}, "b": 4}, "c": True}
    assert c.to_json(sort_keys=True) == json.dumps(expected_result, sort_keys=True)


def test_extract_config_optional():
    @dataclass
    class C(BaseConfig):
        c: bool
        b: t.Optional[B] = None

    c = extract_config(flat_data=flat_data, config=C)
    expected_result = {"b": {"a": {"a": "test"}, "b": 4}, "c": True}
    assert c.to_json(sort_keys=True) == json.dumps(expected_result, sort_keys=True)


def test_extract_config_union():
    @dataclass
    class C(BaseConfig):
        c: bool
        b: t.Optional[t.Union[B, int]] = None

    c = extract_config(flat_data=flat_data, config=C)
    expected_result = {"b": 4, "c": True}
    assert c.to_json(sort_keys=True) == json.dumps(expected_result, sort_keys=True)


def test_extract_config_list():
    @dataclass
    class C(BaseConfig):
        c: t.List[int]
        b: B

    flat_data = {"a": "test", "b": 4, "c": [1, 2, 3]}
    c = extract_config(flat_data=flat_data, config=C)
    expected_result = {"b": {"a": {"a": "test"}, "b": 4}, "c": [1, 2, 3]}
    assert c.to_json(sort_keys=True) == json.dumps(expected_result, sort_keys=True)


def test_extract_config_optional_list():
    @dataclass
    class C(BaseConfig):
        b: B
        c: t.Optional[t.List[int]] = None

    flat_data = {"a": "test", "b": 4, "c": [1, 2, 3]}
    c = extract_config(flat_data=flat_data, config=C)
    expected_result = {"b": {"a": {"a": "test"}, "b": 4}, "c": [1, 2, 3]}
    assert c.to_json(sort_keys=True) == json.dumps(expected_result, sort_keys=True)


def test_extract_config_dataclass_list():
    @dataclass
    class C(BaseConfig):
        c: bool
        b: t.List[B] = field(default_factory=list)

    flat_data = {"a": "test", "c": True}
    c = extract_config(flat_data=flat_data, config=C)
    expected_result = {"b": [], "c": True}
    assert c.to_json(sort_keys=True) == json.dumps(expected_result, sort_keys=True)


def test_extract_config_dict():
    @dataclass
    class C(BaseConfig):
        c: bool
        b: t.Dict[str, B] = field(default_factory=dict)

    flat_data = {"c": True}
    c = extract_config(flat_data=flat_data, config=C)
    expected_result = {"c": True, "b": {}}
    assert c.to_json(sort_keys=True) == json.dumps(expected_result, sort_keys=True)


def test_json_to_dict_valid_json():
    json_string = '{"key": "value"}'
    expected_result = {"key": "value"}
    assert json_to_dict(json_string) == expected_result
    assert isinstance(json_to_dict(json_string), dict)


def test_json_to_dict_malformed_json():
    json_string = '{"key": "value"'
    expected_result = '{"key": "value"'
    assert json_to_dict(json_string) == expected_result
    assert isinstance(json_to_dict(json_string), str)


def test_json_to_dict_single_quotes():
    json_string = "{'key': 'value'}"
    expected_result = {"key": "value"}
    assert json_to_dict(json_string) == expected_result
    assert isinstance(json_to_dict(json_string), dict)


def test_json_to_dict_path():
    json_string = "/path/to/file.json"
    expected_result = "/path/to/file.json"
    assert json_to_dict(json_string) == expected_result
    assert isinstance(json_to_dict(json_string), str)


def test_ensure_isoformat_datetime_for_datetime():
    dt = ensure_isoformat_datetime(datetime(2021, 1, 1, 12, 0, 0))
    assert dt == "2021-01-01T12:00:00"


def test_ensure_isoformat_datetime_for_datetime_with_tz():
    dt = ensure_isoformat_datetime(datetime(2021, 1, 1, 12, 0, 0, tzinfo=pytz.UTC))
    assert dt == "2021-01-01T12:00:00+00:00"


def test_ensure_isoformat_datetime_for_string():
    dt = ensure_isoformat_datetime("2021-01-01T12:00:00")
    assert dt == "2021-01-01T12:00:00"


def test_ensure_isoformat_datetime_for_string2():
    dt = ensure_isoformat_datetime("2021-01-01T12:00:00+00:00")
    assert dt == "2021-01-01T12:00:00+00:00"


def test_ensure_isoformat_datetime_fails_on_string():
    with pytest.raises(ValueError):
        ensure_isoformat_datetime("bad timestamp")


def test_ensure_isoformat_datetime_fails_on_int():
    with pytest.raises(TypeError):
        ensure_isoformat_datetime(1111)


def test_truncate_string_bytes_return_truncated_string():
    test_string = "abcdef안녕하세요ghijklmn방갑습니opqrstu 더 길어지면 안되는 문자열vwxyz"
    max_bytes = 11
    result = truncate_string_bytes(test_string, max_bytes)
    assert result == "abcdef안"
    assert len(result.encode("utf-8")) <= max_bytes


def test_truncate_string_bytes_return_untouched_string():
    test_string = "abcdef"
    max_bytes = 11
    result = truncate_string_bytes(test_string, max_bytes)
    assert result == "abcdef"
    assert len(result.encode("utf-8")) <= max_bytes



from pathlib import Path

import pytest


@pytest.fixture
def embedder_file() -> Path:
    int_test_dir = Path(__file__).parent
    assets_dir = int_test_dir / "assets"
    embedder_file = assets_dir / "DA-1p-with-duplicate-pages.pdf.json"
    assert embedder_file.exists()
    assert embedder_file.is_file()
    return embedder_file



import typing as t
from dataclasses import dataclass

import click

from unstructured_ingest.cli.interfaces import CliConfig
from unstructured_ingest.connector.sql import SimpleSqlConfig
from unstructured_ingest.interfaces import WriteConfig

SQL_DRIVERS = {"postgresql", "sqlite"}


@dataclass
class SqlCliConfig(SimpleSqlConfig, CliConfig):
    @staticmethod
    def get_cli_options() -> t.List[click.Option]:
        options = [
            click.Option(
                ["--db-type"],
                required=True,
                type=click.Choice(SQL_DRIVERS),
                help="Type of the database backend",
            ),
            click.Option(
                ["--username"],
                default=None,
                type=str,
                help="DB username",
            ),
            click.Option(
                ["--password"],
                default=None,
                type=str,
                help="DB password",
            ),
            click.Option(
                ["--host"],
                default=None,
                type=str,
                help="DB host",
            ),
            click.Option(
                ["--port"],
                default=None,
                type=int,
                help="DB host connection port",
            ),
            click.Option(
                ["--database"],
                default=None,
                type=str,
                help="Database name. For sqlite databases, this is the path to the .db file.",
            ),
        ]
        return options


def get_base_dest_cmd():
    from unstructured_ingest.cli.base.dest import BaseDestCmd

    cmd_cls = BaseDestCmd(
        cmd_name="sql",
        cli_config=SqlCliConfig,
        write_config=WriteConfig,
    )
    return cmd_cls



import base64
from pathlib import Path
from typing import TYPE_CHECKING, Optional
from urllib.parse import urlparse
from uuid import NAMESPACE_DNS, uuid5

from pydantic import BaseModel, Field

from unstructured_ingest.utils.dep_check import requires_dependencies
from unstructured_ingest.v2.interfaces import DownloadResponse, FileData, SourceIdentifiers
from unstructured_ingest.v2.logger import logger

if TYPE_CHECKING:
    from bs4.element import Tag
    from requests import Session


class HtmlMixin(BaseModel):
    extract_images: bool = Field(
        default=False,
        description="if true, will download images and replace "
        "the html content with base64 encoded images",
    )
    extract_files: bool = Field(
        default=False, description="if true, will download any embedded files"
    )
    force_download: bool = Field(
        default=False,
        description="if true, will redownload extracted files even if they already exist locally",
    )
    allow_list: Optional[list[str]] = Field(
        default=None,
        description="list of allowed urls to download, if not set, "
        "will default to the base url the original HTML came from",
    )

    @requires_dependencies(["requests"])
    def get_default_session(self) -> "Session":
        import requests

        return requests.Session()

    def get_absolute_url(self, tag_link: str, url: str) -> str:
        parsed_url = urlparse(url)
        base_url = parsed_url.scheme + "://" + parsed_url.netloc
        if tag_link.startswith("//"):
            return f"{parsed_url.scheme}:{tag_link}"
        elif tag_link.startswith("http"):
            return tag_link
        else:
            tag_link = tag_link.lstrip("/")
            return f"{base_url}/{tag_link}"

    def download_content(self, url: str, session: "Session") -> bytes:
        response = session.get(url)
        response.raise_for_status()
        return response.content

    def can_download(self, url_to_download: str, original_url: str) -> bool:
        parsed_original_url = urlparse(original_url)
        base_url = parsed_original_url.scheme + "://" + parsed_original_url.netloc
        allow_list = self.allow_list or [base_url]
        for allowed_url in allow_list:
            if url_to_download.startswith(allowed_url):
                return True
        logger.info(f"Skipping url because it does not match the allow list: {url_to_download}")
        return False

    def extract_image_src(self, image: "Tag", url: str, session: "Session") -> "Tag":
        current_src = image["src"]
        if current_src.startswith("data:image/png;base64"):
            # already base64 encoded
            return image
        absolute_url = self.get_absolute_url(tag_link=image["src"], url=url)
        if not self.can_download(url_to_download=absolute_url, original_url=url):
            return image
        image_content = self.download_content(url=absolute_url, session=session)
        logger.debug("img tag having src updated from {} to base64 content".format(image["src"]))
        image["src"] = f"data:image/png;base64,{base64.b64encode(image_content).decode()}"
        return image

    @requires_dependencies(["bs4"])
    def extract_html_images(self, url: str, html: str, session: Optional["Session"] = None) -> str:
        from bs4 import BeautifulSoup

        session = session or self.get_default_session()
        soup = BeautifulSoup(html, "html.parser")
        images = soup.find_all("img")
        for image in images:
            self.extract_image_src(image=image, url=url, session=session)
        return str(soup)

    @requires_dependencies(["bs4"])
    def get_hrefs(self, url: str, html: str) -> list:
        from bs4 import BeautifulSoup

        soup = BeautifulSoup(html, "html.parser")
        tags = soup.find_all("a", href=True)
        hrefs = [
            tag["href"]
            for tag in tags
            if not tag["href"].startswith("#") and Path(tag["href"]).suffix != ""
        ]
        absolute_urls = [self.get_absolute_url(tag_link=href, url=url) for href in hrefs]
        allowed_urls = [
            url_to_download
            for url_to_download in absolute_urls
            if self.can_download(url_to_download=url_to_download, original_url=url)
        ]
        return allowed_urls

    def write_content(self, content: bytes, path: Path) -> None:
        if path.exists() and path.is_file() and not self.force_download:
            return
        if not path.parent.exists():
            path.parent.mkdir(parents=True)
        with path.open("wb") as f:
            f.write(content)

    def get_download_response(
        self, url: str, download_dir: Path, file_data: FileData, session: "Session"
    ) -> DownloadResponse:
        filename = Path(urlparse(url=url).path).name
        download_path = download_dir / filename
        self.write_content(
            content=self.download_content(url=url, session=session), path=download_path
        )
        result_file_data = file_data.model_copy(deep=True)
        result_file_data.metadata.url = url
        if result_file_data.metadata.record_locator is None:
            result_file_data.metadata.record_locator = {}
        result_file_data.metadata.record_locator["parent_url"] = url
        result_file_data.identifier = str(uuid5(NAMESPACE_DNS, url + file_data.identifier))
        filename = Path(urlparse(url=url).path).name
        result_file_data.source_identifiers = SourceIdentifiers(
            filename=filename, fullpath=filename
        )
        result_file_data.local_download_path = download_path.as_posix()
        return DownloadResponse(file_data=result_file_data, path=download_path)

    def extract_embedded_files(
        self,
        url: str,
        html: str,
        download_dir: Path,
        original_filedata: FileData,
        session: Optional["Session"] = None,
    ) -> list[DownloadResponse]:
        session = session or self.get_default_session()
        urls_to_download = self.get_hrefs(url=url, html=html)
        return [
            self.get_download_response(
                url=url_to_download,
                download_dir=download_dir,
                file_data=original_filedata,
                session=session,
            )
            for url_to_download in urls_to_download
        ]



import typing as t
from dataclasses import dataclass

import click

from unstructured_ingest.cli.base.src import BaseSrcCmd
from unstructured_ingest.cli.interfaces import (
    CliConfig,
)
from unstructured_ingest.connector.wikipedia import SimpleWikipediaConfig


@dataclass
class WikipediaCliConfig(SimpleWikipediaConfig, CliConfig):
    @staticmethod
    def get_cli_options() -> t.List[click.Option]:
        options = [
            click.Option(
                ["--page-title"],
                required=True,
                type=str,
                help='Title of a Wikipedia page, e.g. "Open source software".',
            ),
            click.Option(
                ["--auto-suggest"],
                default=True,
                is_flag=True,
                help="Whether to automatically suggest a page if the exact page was not found."
                " Set to False if the wrong Wikipedia page is fetched.",
            ),
        ]
        return options


def get_base_src_cmd() -> BaseSrcCmd:
    cmd_cls = BaseSrcCmd(
        cmd_name="wikipedia",
        cli_config=WikipediaCliConfig,
    )
    return cmd_cls



import json
import os
import time
from pathlib import Path
from typing import Generator
from uuid import uuid4

import pytest
import requests

from test.integration.connectors.utils.constants import DESTINATION_TAG, NOSQL_TAG
from test.integration.utils import requires_env
from unstructured_ingest.v2.interfaces.file_data import FileData, SourceIdentifiers
from unstructured_ingest.v2.logger import logger
from unstructured_ingest.v2.processes.connectors.vectara import (
    CONNECTOR_TYPE as VECTARA_CONNECTOR_TYPE,
)
from unstructured_ingest.v2.processes.connectors.vectara import (
    VectaraAccessConfig,
    VectaraConnectionConfig,
    VectaraUploader,
    VectaraUploaderConfig,
    VectaraUploadStager,
    VectaraUploadStagerConfig,
)


def validate_upload(response: dict, expected_data: dict):
    element_id = expected_data["element_id"]
    expected_text = expected_data["text"]
    filename = expected_data["metadata"]["filename"]
    filetype = expected_data["metadata"]["filetype"]
    page_number = expected_data["metadata"]["page_number"]

    response = response["search_results"][0]

    assert response is not None
    assert response["text"] == expected_text
    assert response["part_metadata"]["element_id"] == element_id
    assert response["part_metadata"]["filename"] == filename
    assert response["part_metadata"]["filetype"] == filetype
    assert response["part_metadata"]["page_number"] == page_number


@requires_env("VECTARA_OAUTH_CLIENT_ID", "VECTARA_OAUTH_SECRET", "VECTARA_CUSTOMER_ID")
def _get_jwt_token():
    """Connect to the server and get a JWT token."""
    customer_id = os.environ["VECTARA_CUSTOMER_ID"]
    token_endpoint = (
        f"https://vectara-prod-{customer_id}.auth.us-west-2.amazoncognito.com/oauth2/token"
    )
    headers = {
        "Content-Type": "application/x-www-form-urlencoded",
    }
    data = {
        "grant_type": "client_credentials",
        "client_id": os.environ["VECTARA_OAUTH_CLIENT_ID"],
        "client_secret": os.environ["VECTARA_OAUTH_SECRET"],
    }

    response = requests.post(token_endpoint, headers=headers, data=data)
    response.raise_for_status()
    response_json = response.json()

    return response_json.get("access_token")


def query_data(corpus_key: str, element_id: str) -> dict:

    url = f"https://api.vectara.io/v2/corpora/{corpus_key}/query"

    # the query below requires the corpus to have filter attributes for element_id

    data = json.dumps(
        {
            "query": "string",
            "search": {
                "metadata_filter": f"part.element_id = '{element_id}'",
                "lexical_interpolation": 1,
                "limit": 10,
            },
        }
    )

    jwt_token = _get_jwt_token()
    headers = {
        "Content-Type": "application/json",
        "Accept": "application/json",
        "Authorization": f"Bearer {jwt_token}",
        "X-source": "unstructured",
    }

    response = requests.post(url, headers=headers, data=data)
    response.raise_for_status()
    response_json = response.json()

    return response_json


def create_corpora(corpus_key: str, corpus_name: str) -> None:
    url = "https://api.vectara.io/v2/corpora"
    data = json.dumps({"key": corpus_key, "name": corpus_name, "description": "integration test"})
    jwt_token = _get_jwt_token()
    headers = {
        "Content-Type": "application/json",
        "Accept": "application/json",
        "Authorization": f"Bearer {jwt_token}",
        "X-source": "unstructured",
    }

    response = requests.post(url, headers=headers, data=data)
    response.raise_for_status()


def replace_filter_attributes(corpus_key: str) -> None:
    url = f"https://api.vectara.io/v2/corpora/{corpus_key}/replace_filter_attributes"
    data = json.dumps(
        {
            "filter_attributes": [
                {"name": "element_id", "level": "part", "indexed": True, "type": "text"}
            ]
        }
    )
    jwt_token = _get_jwt_token()
    headers = {
        "Content-Type": "application/json",
        "Accept": "application/json",
        "Authorization": f"Bearer {jwt_token}",
        "X-source": "unstructured",
    }

    response = requests.post(url, headers=headers, data=data)
    response.raise_for_status()


def delete_corpora(corpus_key: str) -> None:
    url = f"https://api.vectara.io/v2/corpora/{corpus_key}"

    jwt_token = _get_jwt_token()
    headers = {
        "Content-Type": "application/json",
        "Accept": "application/json",
        "Authorization": f"Bearer {jwt_token}",
        "X-source": "unstructured",
    }

    response = requests.delete(url, headers=headers)
    response.raise_for_status()


def list_corpora() -> list:
    url = "https://api.vectara.io/v2/corpora?limit=100"
    jwt_token = _get_jwt_token()
    headers = {
        "Content-Type": "application/json",
        "Accept": "application/json",
        "Authorization": f"Bearer {jwt_token}",
        "X-source": "unstructured",
    }
    response = requests.get(url, headers=headers)
    response.raise_for_status()
    response_json = response.json()
    if response_json.get("corpora"):
        return [item["key"] for item in response_json.get("corpora")]
    else:
        return []


def wait_for_ready(corpus_key: str, timeout=60, interval=2) -> None:
    def is_ready_status():
        corpora_list = list_corpora()
        return corpus_key in corpora_list

    start = time.time()
    is_ready = is_ready_status()
    while not is_ready and time.time() - start < timeout:
        time.sleep(interval)
        is_ready = is_ready_status()
    if not is_ready:
        raise TimeoutError("time out waiting for corpus to be ready")


def wait_for_delete(corpus_key: str, timeout=60, interval=2) -> None:
    start = time.time()
    while time.time() - start < timeout:
        corpora_list = list_corpora()
        if corpus_key not in corpora_list:
            return
        time.sleep(interval)

    raise TimeoutError("time out waiting for corpus to delete")


@pytest.fixture
def corpora_util() -> Generator[str, None, None]:
    random_id = str(uuid4()).split("-")[0]
    corpus_key = f"ingest-test-{random_id}"
    corpus_name = "ingest-test"
    logger.info(f"Creating corpus with key: {corpus_key}")
    try:
        create_corpora(corpus_key, corpus_name)
        replace_filter_attributes(corpus_key)
        wait_for_ready(corpus_key=corpus_key)
        yield corpus_key
    except Exception as e:
        logger.error(f"failed to create corpus {corpus_key}: {e}")
    finally:
        logger.info(f"deleting corpus: {corpus_key}")
        delete_corpora(corpus_key)
        wait_for_delete(corpus_key=corpus_key)


@pytest.mark.asyncio
@pytest.mark.tags(VECTARA_CONNECTOR_TYPE, DESTINATION_TAG, "vectara", NOSQL_TAG)
@requires_env("VECTARA_OAUTH_CLIENT_ID", "VECTARA_OAUTH_SECRET", "VECTARA_CUSTOMER_ID")
async def test_vectara_destination(
    upload_file: Path, tmp_path: Path, corpora_util: str, retries=30, interval=10
):
    corpus_key = corpora_util
    connection_kwargs = {
        "customer_id": os.environ["VECTARA_CUSTOMER_ID"],
        "corpus_key": corpus_key,
    }

    oauth_client_id = os.environ["VECTARA_OAUTH_CLIENT_ID"]
    oauth_secret = os.environ["VECTARA_OAUTH_SECRET"]

    file_data = FileData(
        source_identifiers=SourceIdentifiers(fullpath=upload_file.name, filename=upload_file.name),
        connector_type=VECTARA_CONNECTOR_TYPE,
        identifier="mock-file-data",
    )

    stager_config = VectaraUploadStagerConfig(batch_size=10)
    stager = VectaraUploadStager(upload_stager_config=stager_config)
    new_upload_file = stager.run(
        elements_filepath=upload_file,
        output_dir=tmp_path,
        output_filename=upload_file.name,
        file_data=file_data,
    )

    uploader = VectaraUploader(
        connection_config=VectaraConnectionConfig(
            **connection_kwargs,
            access_config=VectaraAccessConfig(
                oauth_client_id=oauth_client_id, oauth_secret=oauth_secret
            ),
        ),
        upload_config=VectaraUploaderConfig(),
    )

    with new_upload_file.open() as new_upload_fp:
        elements_stager = json.load(new_upload_fp)

    if uploader.is_async():
        await uploader.run_data_async(data=elements_stager, file_data=file_data)

    with upload_file.open() as upload_fp:
        elements = json.load(upload_fp)
    first_element = elements[0]

    for i in range(retries):
        response = query_data(corpus_key, first_element["element_id"])
        if not response["search_results"]:
            time.sleep(interval)
        else:
            break

    validate_upload(response=response, expected_data=first_element)



import json
from pathlib import Path

import pytest


@pytest.fixture
def collections_schema_config() -> dict:
    int_test_dir = Path(__file__).parent
    assets_dir = int_test_dir / "assets"
    config_file = assets_dir / "elements.json"
    assert config_file.exists()
    assert config_file.is_file()
    with config_file.open() as config_data:
        return json.load(config_data)



import typing as t
from dataclasses import dataclass

import click

from unstructured_ingest.cli.base.src import BaseSrcCmd
from unstructured_ingest.cli.interfaces import (
    CliConfig,
    DelimitedString,
)
from unstructured_ingest.connector.confluence import SimpleConfluenceConfig


@dataclass
class ConfluenceCliConfig(SimpleConfluenceConfig, CliConfig):
    @staticmethod
    def get_cli_options() -> t.List[click.Option]:
        options = [
            click.Option(
                ["--api-token"],
                required=True,
                help="API Token to authenticate into Confluence Cloud. "
                "Check "
                "https://developer.atlassian.com/cloud/confluence/basic-auth-for-rest-apis/ "
                "for more info.",
            ),
            click.Option(
                ["--url"],
                required=True,
                help='URL to Confluence Cloud, e.g. "unstructured-ingest-test.atlassian.net"',
            ),
            click.Option(
                ["--user-email"],
                required=True,
                help="Email to authenticate into Confluence Cloud",
            ),
            click.Option(
                ["--spaces"],
                default=None,
                type=DelimitedString(),
                help="A list of confluence space ids to be fetched. From each fetched space, "
                "--num-of-docs-from-each-space number of docs will be ingested. "
                "--spaces and --num-of-spaces cannot be used at the same time",
            ),
            click.Option(
                ["--max-num-of-docs-from-each-space"],
                default=100,
                help="Number of documents to be aimed to be ingested from each fetched "
                "confluence space. If any space has fewer documents, all the documents from "
                "that space will be ingested. Documents are not necessarily "
                "ingested in order of creation date.",
            ),
            click.Option(
                ["--max-num-of-spaces"],
                default=500,
                help="Number of confluence space ids to be fetched. From each fetched space, "
                "--num-of-docs-from-each-space number of docs will be ingested. "
                "--spaces and --num-of-spaces cannot be used at the same time",
            ),
        ]
        return options


def get_base_src_cmd() -> BaseSrcCmd:
    cmd_cls = BaseSrcCmd(
        cmd_name="confluence",
        cli_config=ConfluenceCliConfig,
    )
    return cmd_cls



import typing as t
from dataclasses import dataclass

from unstructured_ingest.interfaces import BaseSourceConnector
from unstructured_ingest.runner.base_runner import Runner

if t.TYPE_CHECKING:
    from unstructured_ingest.connector.local import SimpleLocalConfig


@dataclass
class LocalRunner(Runner):
    connector_config: "SimpleLocalConfig"

    def update_read_config(self):
        pass

    def get_source_connector_cls(self) -> t.Type[BaseSourceConnector]:
        from unstructured_ingest.connector.local import (
            LocalSourceConnector,
        )

        return LocalSourceConnector



import typing as t
from dataclasses import dataclass

import click

from unstructured_ingest.cli.base.src import BaseSrcCmd
from unstructured_ingest.cli.interfaces import (
    CliConfig,
)
from unstructured_ingest.connector.fsspec.s3 import S3WriteConfig, SimpleS3Config

CMD_NAME = "s3"


@dataclass
class S3CliConfig(SimpleS3Config, CliConfig):
    @staticmethod
    def get_cli_options() -> t.List[click.Option]:
        options = [
            click.Option(
                ["--anonymous"],
                is_flag=True,
                default=False,
                help="Connect to s3 without local AWS credentials.",
            ),
            click.Option(
                ["--endpoint-url"],
                type=str,
                default=None,
                help="Use this endpoint_url, if specified. Needed for "
                "connecting to non-AWS S3 buckets.",
            ),
            click.Option(
                ["--key"],
                type=str,
                default=None,
                help="If not anonymous, use this access key ID, if specified. Takes precedence "
                "over `aws_access_key_id` in client_kwargs.",
            ),
            click.Option(
                ["--secret"],
                type=str,
                default=None,
                help="If not anonymous, use this secret access key, if specified.",
            ),
            click.Option(
                ["--token"],
                type=str,
                default=None,
                help="If not anonymous, use this security token, if specified.",
            ),
        ]
        return options


def get_base_src_cmd():
    cmd_cls = BaseSrcCmd(
        cmd_name=CMD_NAME,
        cli_config=S3CliConfig,
        is_fsspec=True,
    )
    return cmd_cls


def get_base_dest_cmd():
    from unstructured_ingest.cli.base.dest import BaseDestCmd

    cmd_cls = BaseDestCmd(
        cmd_name=CMD_NAME,
        cli_config=S3CliConfig,
        write_config=S3WriteConfig,
        is_fsspec=True,
    )
    return cmd_cls



import typing as t
from dataclasses import dataclass

import click

from unstructured_ingest.cli.base.src import BaseSrcCmd
from unstructured_ingest.cli.interfaces import (
    CliConfig,
    DelimitedString,
)
from unstructured_ingest.connector.discord import SimpleDiscordConfig


@dataclass
class DiscordCliConfig(SimpleDiscordConfig, CliConfig):
    @staticmethod
    def get_cli_options() -> t.List[click.Option]:
        options = [
            click.Option(
                ["--token"],
                required=True,
                help="Bot token used to access Discord API, must have "
                "READ_MESSAGE_HISTORY scope for the bot user",
            ),
            click.Option(
                ["--channels"],
                required=True,
                type=DelimitedString(),
                help="Comma-delimited list of discord channel ids to ingest from.",
            ),
            click.Option(
                ["--period"],
                default=None,
                type=click.IntRange(0),
                help="Number of days to go back in the history of "
                "discord channels, must be a number",
            ),
        ]
        return options


def get_base_src_cmd() -> BaseSrcCmd:
    cmd_cls = BaseSrcCmd(
        cmd_name="discord",
        cli_config=DiscordCliConfig,
    )
    return cmd_cls



import typing as t
from dataclasses import dataclass

from unstructured_ingest.interfaces import BaseDestinationConnector
from unstructured_ingest.runner.writers.base_writer import Writer

if t.TYPE_CHECKING:
    from unstructured_ingest.connector.fsspec.s3 import S3WriteConfig, SimpleS3Config


@dataclass
class S3Writer(Writer):
    connector_config: "SimpleS3Config"
    write_config: "S3WriteConfig"

    def get_connector_cls(self) -> t.Type[BaseDestinationConnector]:
        from unstructured_ingest.connector.fsspec.s3 import (
            S3DestinationConnector,
        )

        return S3DestinationConnector



import typing as t
from dataclasses import dataclass

from unstructured_ingest.connector.fsspec.fsspec import (
    FsspecDestinationConnector,
    FsspecIngestDoc,
    FsspecSourceConnector,
    FsspecWriteConfig,
    SimpleFsspecConfig,
    WriteTextConfig,
)
from unstructured_ingest.enhanced_dataclass import enhanced_field
from unstructured_ingest.error import DestinationConnectionError, SourceConnectionError
from unstructured_ingest.interfaces import AccessConfig
from unstructured_ingest.logger import logger
from unstructured_ingest.utils.dep_check import requires_dependencies


@dataclass
class AzureWriteTextConfig(WriteTextConfig):
    overwrite: bool = False


@dataclass
class AzureWriteConfig(FsspecWriteConfig):
    write_text_config: t.Optional[AzureWriteTextConfig] = None


@dataclass
class AzureAccessConfig(AccessConfig):
    account_name: t.Optional[str] = enhanced_field(default=None, sensitive=True)
    account_key: t.Optional[str] = enhanced_field(default=None, sensitive=True)
    connection_string: t.Optional[str] = enhanced_field(default=None, sensitive=True)
    sas_token: t.Optional[str] = enhanced_field(default=None, sensitive=True)


@dataclass
class SimpleAzureBlobStorageConfig(SimpleFsspecConfig):
    access_config: AzureAccessConfig = None


@dataclass
class AzureBlobStorageIngestDoc(FsspecIngestDoc):
    connector_config: SimpleAzureBlobStorageConfig
    registry_name: str = "azure"

    @SourceConnectionError.wrap
    @requires_dependencies(["adlfs", "fsspec"], extras="azure")
    def get_file(self):
        super().get_file()


@dataclass
class AzureBlobStorageSourceConnector(FsspecSourceConnector):
    connector_config: SimpleAzureBlobStorageConfig

    def __post_init__(self):
        self.ingest_doc_cls: t.Type[AzureBlobStorageIngestDoc] = AzureBlobStorageIngestDoc


@dataclass
class AzureBlobStorageDestinationConnector(FsspecDestinationConnector):
    connector_config: SimpleAzureBlobStorageConfig
    write_config: AzureWriteConfig

    @requires_dependencies(["adlfs", "fsspec"], extras="azure")
    def initialize(self):
        super().initialize()

    @requires_dependencies(["adlfs"], extras="azure")
    def check_connection(self):
        from adlfs import AzureBlobFileSystem

        try:
            AzureBlobFileSystem(**self.connector_config.get_access_config())
        except ValueError as connection_error:
            logger.error(f"failed to validate connection: {connection_error}", exc_info=True)
            raise DestinationConnectionError(f"failed to validate connection: {connection_error}")



import json
from typing import IO, Any


def dumps(obj: list[dict[str, Any]], **kwargs) -> str:
    return "\n".join(json.dumps(each, **kwargs) for each in obj)


def dump(obj: list[dict[str, Any]], fp: IO, **kwargs) -> None:
    # Indent breaks ndjson formatting
    kwargs["indent"] = None
    text = dumps(obj, **kwargs)
    fp.write(text)


def loads(s: str, **kwargs) -> list[dict[str, Any]]:
    return [json.loads(line, **kwargs) for line in s.splitlines()]


def load(fp: IO, **kwargs) -> list[dict[str, Any]]:
    return loads(fp.read(), **kwargs)


class writer(object):
    def __init__(self, f, **kwargs):
        self.f = f
        self.kwargs = kwargs

    def write(self, row):
        stringified = json.dumps(row, **self.kwargs)
        self.f.write(stringified + "\n")


class reader(object):
    def __init__(self, f, **kwargs):
        self.f = f
        self.kwargs = kwargs

    def __iter__(self):
        return self

    def __next__(self):
        line = ""

        while line == "":
            line = next(self.f).strip()

        return json.loads(line, **self.kwargs)

    # NOTE: this is necessary to comply with py27
    def next(self):
        return self.__next__()



from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any


@dataclass
class BaseProcess(ABC):
    def is_async(self) -> bool:
        return False

    def init(self, *kwargs: Any) -> None:
        pass

    def precheck(self) -> None:
        pass

    @abstractmethod
    def run(self, **kwargs: Any) -> Any:
        pass

    async def run_async(self, **kwargs: Any) -> Any:
        return self.run(**kwargs)



import json
import os
import traceback
from dataclasses import dataclass, field
from multiprocessing import Process, Queue
from pathlib import Path
from typing import Any, Optional
from urllib.parse import urlparse

import pandas as pd
from pydantic import Field, Secret

from unstructured_ingest.error import DestinationConnectionError
from unstructured_ingest.utils.data_prep import get_data_df
from unstructured_ingest.utils.dep_check import requires_dependencies
from unstructured_ingest.utils.table import convert_to_pandas_dataframe
from unstructured_ingest.v2.interfaces import (
    AccessConfig,
    ConnectionConfig,
    FileData,
    Uploader,
    UploaderConfig,
    UploadStager,
    UploadStagerConfig,
)
from unstructured_ingest.v2.logger import logger
from unstructured_ingest.v2.processes.connector_registry import DestinationRegistryEntry

CONNECTOR_TYPE = "delta_table"


@requires_dependencies(["deltalake"], extras="delta-table")
def write_deltalake_with_error_handling(queue, **kwargs):
    from deltalake.writer import write_deltalake

    try:
        write_deltalake(**kwargs)
    except Exception:
        queue.put(traceback.format_exc())


class DeltaTableAccessConfig(AccessConfig):
    aws_access_key_id: Optional[str] = Field(default=None, description="AWS Access Key Id")
    aws_secret_access_key: Optional[str] = Field(default=None, description="AWS Secret Access Key")


class DeltaTableConnectionConfig(ConnectionConfig):
    access_config: Secret[DeltaTableAccessConfig] = Field(
        default=DeltaTableAccessConfig(), validate_default=True
    )
    aws_region: Optional[str] = Field(default=None, description="AWS Region")
    table_uri: str = Field(
        default=None,
        description=(
            "Local path or path to the target folder in the S3 bucket, "
            "formatted as s3://my-bucket/my-folder/"
        ),
    )

    def update_storage_options(self, storage_options: dict) -> None:
        secrets = self.access_config.get_secret_value()
        if self.aws_region and secrets.aws_access_key_id and secrets.aws_secret_access_key:
            storage_options["AWS_REGION"] = self.aws_region
            storage_options["AWS_ACCESS_KEY_ID"] = secrets.aws_access_key_id
            storage_options["AWS_SECRET_ACCESS_KEY"] = secrets.aws_secret_access_key
            # Delta-rs doesn't support concurrent S3 writes without external locks (DynamoDB).
            # This flag allows single-writer uploads to S3 without using locks, according to:
            # https://delta-io.github.io/delta-rs/usage/writing/writing-to-s3-with-locking-provider/
            storage_options["AWS_S3_ALLOW_UNSAFE_RENAME"] = "true"


class DeltaTableUploadStagerConfig(UploadStagerConfig):
    pass


@dataclass
class DeltaTableUploadStager(UploadStager):
    upload_stager_config: DeltaTableUploadStagerConfig = field(
        default_factory=lambda: DeltaTableUploadStagerConfig()
    )

    def run(
        self,
        elements_filepath: Path,
        output_dir: Path,
        output_filename: str,
        **kwargs: Any,
    ) -> Path:
        with open(elements_filepath) as elements_file:
            elements_contents = json.load(elements_file)

        output_path = Path(output_dir) / Path(f"{output_filename}.parquet")

        df = convert_to_pandas_dataframe(elements_dict=elements_contents)
        df.to_parquet(output_path)

        return output_path


class DeltaTableUploaderConfig(UploaderConfig):
    pass


@dataclass
class DeltaTableUploader(Uploader):
    upload_config: DeltaTableUploaderConfig
    connection_config: DeltaTableConnectionConfig
    connector_type: str = CONNECTOR_TYPE

    @requires_dependencies(["boto3"], extras="delta-table")
    def precheck(self):
        secrets = self.connection_config.access_config.get_secret_value()
        if (
            self.connection_config.aws_region
            and secrets.aws_access_key_id
            and secrets.aws_secret_access_key
        ):
            from boto3 import client

            url = urlparse(self.connection_config.table_uri)
            bucket_name = url.netloc
            dir_path = url.path.lstrip("/")

            try:
                s3_client = client(
                    "s3",
                    aws_access_key_id=secrets.aws_access_key_id,
                    aws_secret_access_key=secrets.aws_secret_access_key,
                )
                s3_client.put_object(Bucket=bucket_name, Key=dir_path, Body=b"")

                response = s3_client.get_bucket_location(Bucket=bucket_name)

                if self.connection_config.aws_region != response.get("LocationConstraint"):
                    raise ValueError("Wrong AWS Region was provided.")

            except Exception as e:
                logger.error(f"failed to validate connection: {e}", exc_info=True)
                raise DestinationConnectionError(f"failed to validate connection: {e}")

    def upload_dataframe(self, df: pd.DataFrame, file_data: FileData) -> None:
        updated_upload_path = os.path.join(
            self.connection_config.table_uri, file_data.source_identifiers.relative_path
        )
        logger.info(
            f"writing {len(df)} rows to destination table "
            f"at {updated_upload_path}\ndtypes: {df.dtypes}",
        )
        storage_options = {}
        self.connection_config.update_storage_options(storage_options=storage_options)

        writer_kwargs = {
            "table_or_uri": updated_upload_path,
            "data": df,
            "mode": "overwrite",
            "storage_options": storage_options,
        }
        queue = Queue()
        # NOTE: deltalake writer on Linux sometimes can finish but still trigger a SIGABRT and cause
        # ingest to fail, even though all tasks are completed normally. Putting the writer into a
        # process mitigates this issue by ensuring python interpreter waits properly for deltalake's
        # rust backend to finish
        writer = Process(
            target=write_deltalake_with_error_handling,
            kwargs={"queue": queue, **writer_kwargs},
        )
        writer.start()
        writer.join()

        # Check if the queue has any exception message
        if not queue.empty():
            error_message = queue.get()
            logger.error(f"Exception occurred in write_deltalake: {error_message}")
            raise RuntimeError(f"Error in write_deltalake: {error_message}")

    def run_data(self, data: list[dict], file_data: FileData, **kwargs: Any) -> None:
        df = pd.DataFrame(data=data)
        self.upload_dataframe(df=df, file_data=file_data)

    def run(self, path: Path, file_data: FileData, **kwargs: Any) -> None:
        df = get_data_df(path)
        self.upload_dataframe(df=df, file_data=file_data)


delta_table_destination_entry = DestinationRegistryEntry(
    connection_config=DeltaTableConnectionConfig,
    uploader=DeltaTableUploader,
    uploader_config=DeltaTableUploaderConfig,
    upload_stager=DeltaTableUploadStager,
    upload_stager_config=DeltaTableUploadStagerConfig,
)



import typing as t
from dataclasses import dataclass

import click

from unstructured_ingest.cli.base.src import BaseSrcCmd
from unstructured_ingest.cli.interfaces import (
    CliConfig,
    CliRecursiveConfig,
)
from unstructured_ingest.connector.sharepoint import SimpleSharepointConfig


@dataclass
class SharepointCliConfig(SimpleSharepointConfig, CliConfig):
    @staticmethod
    def get_cli_options() -> t.List[click.Option]:
        options = [
            click.Option(
                ["--client-id"],
                default=None,
                type=str,
                help="Sharepoint app client ID",
            ),
            click.Option(
                ["--client-cred"],
                default=None,
                type=str,
                help="Sharepoint app secret",
            ),
            click.Option(
                ["--site"],
                default=None,
                type=str,
                help="Sharepoint site url. Process either base url e.g \
                    https://[tenant].sharepoint.com  or relative sites \
                    https://[tenant].sharepoint.com/sites/<site_name>. \
                    To process all sites within the tenant pass a site url as \
                    https://[tenant]-admin.sharepoint.com.\
                    This requires the app to be registered at a tenant level",
            ),
            click.Option(
                ["--path"],
                default="Shared Documents",
                type=str,
                help="Path from which to start parsing files. If the connector is to \
                process all sites  within the tenant this filter will be applied to \
                all sites document libraries. Default 'Shared Documents'",
            ),
            click.Option(
                ["--files-only"],
                is_flag=True,
                default=False,
                help="Process only files.",
            ),
        ]
        return options


def get_base_src_cmd() -> BaseSrcCmd:
    cmd_cls = BaseSrcCmd(
        cmd_name="sharepoint",
        cli_config=SharepointCliConfig,
        additional_cli_options=[CliRecursiveConfig],
    )
    return cmd_cls



from __future__ import annotations

from unstructured_ingest.v2.processes.connector_registry import add_destination_entry

from .aws import CONNECTOR_TYPE as LANCEDB_S3_CONNECTOR_TYPE
from .aws import lancedb_aws_destination_entry
from .azure import CONNECTOR_TYPE as LANCEDB_AZURE_CONNECTOR_TYPE
from .azure import lancedb_azure_destination_entry
from .cloud import CONNECTOR_TYPE as LANCEDB_CLOUD_CONNECTOR_TYPE
from .cloud import lancedb_cloud_destination_entry
from .gcp import CONNECTOR_TYPE as LANCEDB_GCS_CONNECTOR_TYPE
from .gcp import lancedb_gcp_destination_entry
from .local import CONNECTOR_TYPE as LANCEDB_LOCAL_CONNECTOR_TYPE
from .local import lancedb_local_destination_entry

add_destination_entry(
    destination_type=LANCEDB_S3_CONNECTOR_TYPE, entry=lancedb_aws_destination_entry
)
add_destination_entry(
    destination_type=LANCEDB_AZURE_CONNECTOR_TYPE, entry=lancedb_azure_destination_entry
)
add_destination_entry(
    destination_type=LANCEDB_GCS_CONNECTOR_TYPE, entry=lancedb_gcp_destination_entry
)
add_destination_entry(
    destination_type=LANCEDB_LOCAL_CONNECTOR_TYPE, entry=lancedb_local_destination_entry
)
add_destination_entry(
    destination_type=LANCEDB_CLOUD_CONNECTOR_TYPE, entry=lancedb_cloud_destination_entry
)



import random

import faker
import pytest

from unstructured_ingest.v2.processes.chunker import Chunker, ChunkerConfig

fake = faker.Faker()


def generate_chunker_config_params() -> dict:
    params = {}
    random_val = random.random()
    if random_val < 0.5:
        params["chunking_strategy"] = fake.word() if random.random() < 0.5 else None
        params["chunk_combine_text_under_n_chars"] = (
            fake.random_int() if random.random() < 0.5 else None
        )
        params["chunk_include_orig_elements"] = fake.boolean() if random.random() < 0.5 else None
        params["chunk_max_characters"] = fake.random_int()
        params["chunk_multipage_sections"] = fake.boolean()
        params["chunk_new_after_n_chars"] = fake.random_int() if random.random() < 0.5 else None
        params["chunk_overlap"] = fake.random_int() if random.random() < 0.5 else None
        params["chunk_overlap_all"] = fake.boolean() if random.random() < 0.5 else None
    if random_val < 0.5:
        params["chunk_by_api"] = True
        params["chunking_endpoint"] = fake.url()
        params["chunk_api_key"] = fake.password()
    else:
        params["chunk_by_api"] = False

    return params


@pytest.mark.parametrize(
    "partition_config_params", [generate_chunker_config_params() for i in range(10)]
)
def test_chunker_config(partition_config_params: dict):
    chunker_config = ChunkerConfig.model_validate(partition_config_params)
    assert chunker_config


@pytest.mark.parametrize(
    "partition_config_params", [generate_chunker_config_params() for i in range(10)]
)
def test_chunker(partition_config_params: dict):
    chunker_config = ChunkerConfig.model_validate(partition_config_params)
    chunker = Chunker(config=chunker_config)
    assert chunker



import typing as t
from dataclasses import dataclass

from unstructured_ingest.enhanced_dataclass import EnhancedDataClassJsonMixin
from unstructured_ingest.interfaces import BaseDestinationConnector
from unstructured_ingest.runner.writers.base_writer import Writer

if t.TYPE_CHECKING:
    from unstructured_ingest.connector.databricks_volumes import (
        DatabricksVolumesWriteConfig,
        SimpleDatabricksVolumesConfig,
    )


@dataclass
class DatabricksVolumesWriter(Writer, EnhancedDataClassJsonMixin):
    write_config: "DatabricksVolumesWriteConfig"
    connector_config: "SimpleDatabricksVolumesConfig"

    def get_connector_cls(self) -> t.Type[BaseDestinationConnector]:
        from unstructured_ingest.connector.databricks_volumes import (
            DatabricksVolumesDestinationConnector,
        )

        return DatabricksVolumesDestinationConnector



# https://developers.notion.com/reference/property-object#select
from dataclasses import dataclass, field
from typing import List, Optional

from htmlBuilder.attributes import Style
from htmlBuilder.tags import Div, HtmlTag

from unstructured_ingest.connector.notion.interfaces import (
    DBCellBase,
    DBPropertyBase,
    FromJSONMixin,
)


@dataclass
class SelectOption(FromJSONMixin):
    color: str
    id: str
    name: str

    @classmethod
    def from_dict(cls, data: dict):
        return cls(**data)


@dataclass
class SelectProp(FromJSONMixin):
    options: List[SelectOption] = field(default_factory=list)

    @classmethod
    def from_dict(cls, data: dict):
        return cls(options=[SelectOption.from_dict(o) for o in data.get("options", [])])


@dataclass
class Select(DBPropertyBase):
    id: str
    name: str
    select: SelectProp
    type: str = "select"

    @classmethod
    def from_dict(cls, data: dict):
        return cls(select=SelectProp.from_dict(data.pop("select", {})), **data)


@dataclass
class SelectCell(DBCellBase):
    id: str
    select: Optional[SelectOption]
    type: str = "select"
    name: Optional[str] = None

    @classmethod
    def from_dict(cls, data: dict):
        select_data = data.pop("select")
        select = None
        if select_data:
            select = SelectOption.from_dict(select_data)
        return cls(select=select, **data)

    def get_html(self) -> Optional[HtmlTag]:
        if select := self.select:
            select_attr = []
            if select.color and select.color != "default":
                select_attr.append(Style(f"color: {select.color}"))
            return Div(select_attr, select.name)
        return None



import asyncio
import hashlib
from dataclasses import dataclass
from pathlib import Path
from typing import Callable, Optional, TypedDict

from unstructured_ingest.utils.data_prep import write_data
from unstructured_ingest.v2.interfaces import FileData
from unstructured_ingest.v2.interfaces.file_data import file_data_from_file
from unstructured_ingest.v2.logger import logger
from unstructured_ingest.v2.pipeline.interfaces import PipelineStep
from unstructured_ingest.v2.processes.partitioner import Partitioner
from unstructured_ingest.v2.utils import serialize_base_model_json

STEP_ID = "partition"


class PartitionStepResponse(TypedDict):
    file_data_path: str
    path: str


@dataclass
class PartitionStep(PipelineStep):
    process: Partitioner
    identifier: str = STEP_ID

    def __str__(self):
        return f"{self.identifier} ({self.process.config.strategy})"

    def __post_init__(self):
        config = self.process.config.model_dump_json()
        logger.info(f"created {self.identifier} with configs: {config}")

    def should_partition(self, filepath: Path, file_data: FileData) -> bool:
        if self.context.reprocess or file_data.reprocess:
            return True
        return not filepath.exists()

    def get_output_filepath(self, filename: Path) -> Path:
        hashed_output_file = f"{self.get_hash(extras=[filename.name])}.json"
        filepath = (self.cache_dir / hashed_output_file).resolve()
        filepath.parent.mkdir(parents=True, exist_ok=True)
        return filepath

    def _save_output(self, output_filepath: str, partitioned_content: list[dict]):
        logger.debug(f"writing partitioned output to: {output_filepath}")
        write_data(path=Path(output_filepath), data=partitioned_content)

    async def _run_async(
        self, fn: Callable, path: str, file_data_path: str
    ) -> Optional[PartitionStepResponse]:
        path = Path(path)
        file_data = file_data_from_file(path=file_data_path)
        output_filepath = self.get_output_filepath(filename=Path(file_data_path))
        if not self.should_partition(filepath=output_filepath, file_data=file_data):
            logger.debug(f"skipping partitioning, output already exists: {output_filepath}")
            return PartitionStepResponse(file_data_path=file_data_path, path=str(output_filepath))
        fn_kwargs = {"filename": path, "metadata": file_data.metadata.model_dump()}
        if not asyncio.iscoroutinefunction(fn):
            partitioned_content = fn(**fn_kwargs)
        elif semaphore := self.context.semaphore:
            async with semaphore:
                partitioned_content = await fn(**fn_kwargs)
        else:
            partitioned_content = await fn(**fn_kwargs)
        self._save_output(
            output_filepath=str(output_filepath), partitioned_content=partitioned_content
        )
        return PartitionStepResponse(file_data_path=file_data_path, path=str(output_filepath))

    def get_hash(self, extras: Optional[list[str]]) -> str:
        hashable_string = serialize_base_model_json(
            model=self.process.config, sort_keys=True, ensure_ascii=True
        )
        if extras:
            hashable_string += "".join(extras)
        return hashlib.sha256(hashable_string.encode()).hexdigest()[:12]



from typing import Any, Generator, List, Optional, Tuple

import notion_client.errors
from notion_client import Client as NotionClient
from notion_client.api_endpoints import BlocksChildrenEndpoint as NotionBlocksChildrenEndpoint
from notion_client.api_endpoints import BlocksEndpoint as NotionBlocksEndpoint
from notion_client.api_endpoints import DatabasesEndpoint as NotionDatabasesEndpoint
from notion_client.api_endpoints import Endpoint
from notion_client.api_endpoints import PagesEndpoint as NotionPagesEndpoint
from notion_client.errors import RequestTimeoutError

from unstructured_ingest.connector.notion.types.block import Block
from unstructured_ingest.connector.notion.types.database import Database
from unstructured_ingest.connector.notion.types.database_properties import (
    map_cells,
)
from unstructured_ingest.connector.notion.types.page import Page
from unstructured_ingest.ingest_backoff import RetryHandler
from unstructured_ingest.interfaces import RetryStrategyConfig
from unstructured_ingest.utils.dep_check import requires_dependencies


@requires_dependencies(["httpx"], extras="notion")
def _get_retry_strategy(
    endpoint: Endpoint, retry_strategy_config: RetryStrategyConfig
) -> RetryHandler:
    import backoff
    import httpx

    retryable_exceptions = (
        httpx.TimeoutException,
        httpx.HTTPStatusError,
        notion_client.errors.HTTPResponseError,
    )

    return RetryHandler(
        backoff.expo,
        retryable_exceptions,
        max_time=retry_strategy_config.max_retry_time,
        max_tries=retry_strategy_config.max_retries,
        logger=endpoint.parent.logger,
        start_log_level=endpoint.parent.logger.level,
        backoff_log_level=endpoint.parent.logger.level,
    )


def get_retry_handler(endpoint: Endpoint) -> Optional[RetryHandler]:
    if retry_strategy_config := getattr(endpoint, "retry_strategy_config"):
        return _get_retry_strategy(endpoint=endpoint, retry_strategy_config=retry_strategy_config)
    return None


class BlocksChildrenEndpoint(NotionBlocksChildrenEndpoint):
    def __init__(
        self,
        *args,
        retry_strategy_config: Optional[RetryStrategyConfig] = None,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.retry_strategy_config = retry_strategy_config

    @property
    def retry_handler(self) -> Optional[RetryHandler]:
        return get_retry_handler(self)

    def list(self, block_id: str, **kwargs: Any) -> Tuple[List[Block], dict]:
        resp: dict = (
            self.retry_handler(super().list, block_id=block_id, **kwargs)
            if self.retry_handler
            else super().list(block_id=block_id, **kwargs)
        )  # type: ignore
        child_blocks = [Block.from_dict(data=b) for b in resp.pop("results", [])]
        return child_blocks, resp

    def iterate_list(
        self,
        block_id: str,
        **kwargs: Any,
    ) -> Generator[List[Block], None, None]:
        while True:
            response: dict = (
                self.retry_handler(super().list, block_id=block_id, **kwargs)
                if self.retry_handler
                else super().list(block_id=block_id, **kwargs)
            )  # type: ignore
            child_blocks = [Block.from_dict(data=b) for b in response.pop("results", [])]
            yield child_blocks

            next_cursor = response.get("next_cursor")
            if not response.get("has_more") or not next_cursor:
                return


class DatabasesEndpoint(NotionDatabasesEndpoint):
    def __init__(
        self,
        *args,
        retry_strategy_config: Optional[RetryStrategyConfig] = None,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.retry_strategy_config = retry_strategy_config

    @property
    def retry_handler(self) -> Optional[RetryHandler]:
        return get_retry_handler(self)

    def retrieve(self, database_id: str, **kwargs: Any) -> Database:
        resp: dict = (
            self.retry_handler(super().retrieve, database_id=database_id, **kwargs)
            if (self.retry_handler)
            else (super().retrieve(database_id=database_id, **kwargs))
        )  # type: ignore
        return Database.from_dict(data=resp)

    @requires_dependencies(["httpx"], extras="notion")
    def retrieve_status(self, database_id: str, **kwargs) -> int:
        import httpx

        request = self.parent._build_request(
            method="HEAD",
            path=f"databases/{database_id}",
            auth=kwargs.get("auth"),
        )
        try:
            response: httpx.Response = (
                self.retry_handler(self.parent.client.send, request)
                if (self.retry_handler)
                else (self.parent.client.send(request))
            )  # type: ignore
            return response.status_code
        except httpx.TimeoutException:
            raise RequestTimeoutError()

    def query(self, database_id: str, **kwargs: Any) -> Tuple[List[Page], dict]:
        """Get a list of [Pages](https://developers.notion.com/reference/page) contained in the database.

        *[🔗 Endpoint documentation](https://developers.notion.com/reference/post-database-query)*
        """  # noqa: E501
        resp: dict = (
            self.retry_handler(super().query, database_id=database_id, **kwargs)
            if (self.retry_handler)
            else (super().query(database_id=database_id, **kwargs))
        )  # type: ignore
        pages = [Page.from_dict(data=p) for p in resp.pop("results")]
        for p in pages:
            p.properties = map_cells(p.properties)
        return pages, resp

    def iterate_query(self, database_id: str, **kwargs: Any) -> Generator[List[Page], None, None]:
        while True:
            response: dict = (
                self.retry_handler(super().query, database_id=database_id, **kwargs)
                if (self.retry_handler)
                else (super().query(database_id=database_id, **kwargs))
            )  # type: ignore
            pages = [Page.from_dict(data=p) for p in response.pop("results", [])]
            for p in pages:
                p.properties = map_cells(p.properties)
            yield pages

            next_cursor = response.get("next_cursor")
            if not response.get("has_more") or not next_cursor:
                return


class BlocksEndpoint(NotionBlocksEndpoint):
    def __init__(
        self,
        *args: Any,
        retry_strategy_config: Optional[RetryStrategyConfig] = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.retry_strategy_config = retry_strategy_config
        self.children = BlocksChildrenEndpoint(
            retry_strategy_config=retry_strategy_config,
            *args,
            **kwargs,
        )

    @property
    def retry_handler(self) -> Optional[RetryHandler]:
        return get_retry_handler(self)

    def retrieve(self, block_id: str, **kwargs: Any) -> Block:
        resp: dict = (
            self.retry_handler(super().retrieve, block_id=block_id, **kwargs)
            if (self.retry_handler)
            else (super().retrieve(block_id=block_id, **kwargs))
        )  # type: ignore
        return Block.from_dict(data=resp)


class PagesEndpoint(NotionPagesEndpoint):
    def __init__(
        self,
        *args,
        retry_strategy_config: Optional[RetryStrategyConfig] = None,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.retry_strategy_config = retry_strategy_config

    @property
    def retry_handler(self) -> Optional[RetryHandler]:
        return get_retry_handler(self)

    def retrieve(self, page_id: str, **kwargs: Any) -> Page:
        resp: dict = (
            self.retry_handler(super().retrieve, page_id=page_id, **kwargs)
            if (self.retry_handler)
            else (super().retrieve(page_id=page_id, **kwargs))
        )  # type: ignore
        return Page.from_dict(data=resp)

    @requires_dependencies(["httpx"], extras="notion")
    def retrieve_status(self, page_id: str, **kwargs) -> int:
        import httpx

        request = self.parent._build_request(
            method="HEAD",
            path=f"pages/{page_id}",
            auth=kwargs.get("auth"),
        )
        try:
            response: httpx.Response = (
                self.retry_handler(self.parent.client.send, request)
                if (self.retry_handler)
                else (self.parent.client.send(request))
            )  # type: ignore
            return response.status_code
        except httpx.TimeoutException:
            raise RequestTimeoutError()


class Client(NotionClient):
    def __init__(
        self,
        *args: Any,
        retry_strategy_config: Optional[RetryStrategyConfig] = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.blocks = BlocksEndpoint(retry_strategy_config=retry_strategy_config, parent=self)
        self.pages = PagesEndpoint(retry_strategy_config=retry_strategy_config, parent=self)
        self.databases = DatabasesEndpoint(retry_strategy_config=retry_strategy_config, parent=self)



from dataclasses import dataclass, field
from typing import Optional

from pydantic import Field, Secret

from unstructured_ingest.v2.processes.connector_registry import (
    DestinationRegistryEntry,
    SourceRegistryEntry,
)
from unstructured_ingest.v2.processes.connectors.databricks.volumes import (
    DatabricksVolumesAccessConfig,
    DatabricksVolumesConnectionConfig,
    DatabricksVolumesDownloader,
    DatabricksVolumesDownloaderConfig,
    DatabricksVolumesIndexer,
    DatabricksVolumesIndexerConfig,
    DatabricksVolumesUploader,
    DatabricksVolumesUploaderConfig,
)

CONNECTOR_TYPE = "databricks_volumes_gcp"


class DatabricksGoogleVolumesAccessConfig(DatabricksVolumesAccessConfig):
    account_id: Optional[str] = Field(
        default=None,
        description="The Databricks account ID for the Databricks " "accounts endpoint.",
    )
    profile: Optional[str] = None
    google_credentials: Optional[str] = None
    google_service_account: Optional[str] = None


class DatabricksGoogleVolumesConnectionConfig(DatabricksVolumesConnectionConfig):
    access_config: Secret[DatabricksGoogleVolumesAccessConfig]


class DatabricksGoogleVolumesIndexerConfig(DatabricksVolumesIndexerConfig):
    pass


@dataclass
class DatabricksGoogleVolumesIndexer(DatabricksVolumesIndexer):
    connection_config: DatabricksGoogleVolumesConnectionConfig
    index_config: DatabricksGoogleVolumesIndexerConfig
    connector_type: str = CONNECTOR_TYPE


class DatabricksGoogleVolumesDownloaderConfig(DatabricksVolumesDownloaderConfig):
    pass


@dataclass
class DatabricksGoogleVolumesDownloader(DatabricksVolumesDownloader):
    connection_config: DatabricksGoogleVolumesConnectionConfig
    download_config: DatabricksVolumesDownloaderConfig
    connector_type: str = CONNECTOR_TYPE


class DatabricksGoogleVolumesUploaderConfig(DatabricksVolumesUploaderConfig):
    pass


@dataclass
class DatabricksGoogleVolumesUploader(DatabricksVolumesUploader):
    connection_config: DatabricksGoogleVolumesConnectionConfig
    upload_config: DatabricksGoogleVolumesUploaderConfig = field(
        default_factory=DatabricksGoogleVolumesUploaderConfig
    )
    connector_type: str = CONNECTOR_TYPE


databricks_gcp_volumes_destination_entry = DestinationRegistryEntry(
    connection_config=DatabricksGoogleVolumesConnectionConfig,
    uploader=DatabricksGoogleVolumesUploader,
    uploader_config=DatabricksGoogleVolumesUploaderConfig,
)

databricks_gcp_volumes_source_entry = SourceRegistryEntry(
    connection_config=DatabricksGoogleVolumesConnectionConfig,
    indexer=DatabricksGoogleVolumesIndexer,
    indexer_config=DatabricksGoogleVolumesIndexerConfig,
    downloader=DatabricksGoogleVolumesDownloader,
    downloader_config=DatabricksGoogleVolumesDownloaderConfig,
)



from __future__ import annotations

import collections
import typing as t

from unstructured_ingest.cli.base.src import BaseSrcCmd
from unstructured_ingest.cli.cmds.fsspec.sftp import get_base_src_cmd as sftp_base_src_cmd

from .airtable import get_base_src_cmd as airtable_base_src_cmd
from .astradb import get_base_dest_cmd as astradb_base_dest_cmd
from .astradb import get_base_src_cmd as astradb_base_src_cmd
from .azure_ai_search import get_base_dest_cmd as azure_ai_search_base_dest_cmd
from .biomed import get_base_src_cmd as biomed_base_src_cmd
from .chroma import get_base_dest_cmd as chroma_base_dest_cmd
from .clarifai import get_base_dest_cmd as clarifai_base_dest_cmd
from .confluence import get_base_src_cmd as confluence_base_src_cmd
from .databricks_volumes import get_base_dest_cmd as databricks_volumes_dest_cmd
from .delta_table import get_base_dest_cmd as delta_table_dest_cmd
from .delta_table import get_base_src_cmd as delta_table_base_src_cmd
from .discord import get_base_src_cmd as discord_base_src_cmd
from .elasticsearch import get_base_dest_cmd as elasticsearch_base_dest_cmd
from .elasticsearch import get_base_src_cmd as elasticsearch_base_src_cmd
from .fsspec.azure import get_base_dest_cmd as azure_base_dest_cmd
from .fsspec.azure import get_base_src_cmd as azure_base_src_cmd
from .fsspec.box import get_base_dest_cmd as box_base_dest_cmd
from .fsspec.box import get_base_src_cmd as box_base_src_cmd
from .fsspec.dropbox import get_base_dest_cmd as dropbox_base_dest_cmd
from .fsspec.dropbox import get_base_src_cmd as dropbox_base_src_cmd
from .fsspec.fsspec import get_base_dest_cmd as fsspec_base_dest_cmd
from .fsspec.fsspec import get_base_src_cmd as fsspec_base_src_cmd
from .fsspec.gcs import get_base_dest_cmd as gcs_base_dest_cmd
from .fsspec.gcs import get_base_src_cmd as gcs_base_src_cmd
from .fsspec.s3 import get_base_dest_cmd as s3_base_dest_cmd
from .fsspec.s3 import get_base_src_cmd as s3_base_src_cmd
from .github import get_base_src_cmd as github_base_src_cmd
from .gitlab import get_base_src_cmd as gitlab_base_src_cmd
from .google_drive import get_base_src_cmd as google_drive_base_src_cmd
from .hubspot import get_base_src_cmd as hubspot_base_src_cmd
from .jira import get_base_src_cmd as jira_base_src_cmd
from .kafka import get_base_dest_cmd as kafka_base_dest_cmd
from .kafka import get_base_src_cmd as kafka_base_src_cmd
from .local import get_base_src_cmd as local_base_src_cmd
from .mongodb import get_base_dest_cmd as mongo_base_dest_cmd
from .mongodb import get_base_src_cmd as mongodb_base_src_cmd
from .notion import get_base_src_cmd as notion_base_src_cmd
from .onedrive import get_base_src_cmd as onedrive_base_src_cmd
from .opensearch import get_base_dest_cmd as opensearch_base_dest_cmd
from .opensearch import get_base_src_cmd as opensearch_base_src_cmd
from .outlook import get_base_src_cmd as outlook_base_src_cmd
from .pinecone import get_base_dest_cmd as pinecone_base_dest_cmd
from .qdrant import get_base_dest_cmd as qdrant_base_dest_cmd
from .reddit import get_base_src_cmd as reddit_base_src_cmd
from .salesforce import get_base_src_cmd as salesforce_base_src_cmd
from .sharepoint import get_base_src_cmd as sharepoint_base_src_cmd
from .slack import get_base_src_cmd as slack_base_src_cmd
from .sql import get_base_dest_cmd as sql_base_dest_cmd
from .vectara import get_base_dest_cmd as vectara_base_dest_cmd
from .weaviate import get_base_dest_cmd as weaviate_dest_cmd
from .wikipedia import get_base_src_cmd as wikipedia_base_src_cmd

if t.TYPE_CHECKING:
    from unstructured_ingest.cli.base.dest import BaseDestCmd

base_src_cmd_fns: t.List[t.Callable[[], BaseSrcCmd]] = [
    airtable_base_src_cmd,
    astradb_base_src_cmd,
    azure_base_src_cmd,
    biomed_base_src_cmd,
    box_base_src_cmd,
    confluence_base_src_cmd,
    delta_table_base_src_cmd,
    discord_base_src_cmd,
    dropbox_base_src_cmd,
    elasticsearch_base_src_cmd,
    fsspec_base_src_cmd,
    gcs_base_src_cmd,
    github_base_src_cmd,
    gitlab_base_src_cmd,
    google_drive_base_src_cmd,
    hubspot_base_src_cmd,
    jira_base_src_cmd,
    kafka_base_src_cmd,
    local_base_src_cmd,
    mongodb_base_src_cmd,
    notion_base_src_cmd,
    onedrive_base_src_cmd,
    opensearch_base_src_cmd,
    outlook_base_src_cmd,
    reddit_base_src_cmd,
    salesforce_base_src_cmd,
    sftp_base_src_cmd,
    sharepoint_base_src_cmd,
    slack_base_src_cmd,
    s3_base_src_cmd,
    wikipedia_base_src_cmd,
]

# Make sure there are not overlapping names
src_cmd_names = [b().cmd_name for b in base_src_cmd_fns]
src_duplicates = [item for item, count in collections.Counter(src_cmd_names).items() if count > 1]
if src_duplicates:
    raise ValueError(
        "multiple base src commands defined with the same names: {}".format(
            ", ".join(src_duplicates),
        ),
    )

base_dest_cmd_fns: t.List[t.Callable[[], "BaseDestCmd"]] = [
    astradb_base_dest_cmd,
    azure_base_dest_cmd,
    box_base_dest_cmd,
    chroma_base_dest_cmd,
    clarifai_base_dest_cmd,
    databricks_volumes_dest_cmd,
    dropbox_base_dest_cmd,
    elasticsearch_base_dest_cmd,
    fsspec_base_dest_cmd,
    gcs_base_dest_cmd,
    kafka_base_dest_cmd,
    s3_base_dest_cmd,
    azure_ai_search_base_dest_cmd,
    delta_table_dest_cmd,
    sql_base_dest_cmd,
    weaviate_dest_cmd,
    mongo_base_dest_cmd,
    pinecone_base_dest_cmd,
    qdrant_base_dest_cmd,
    opensearch_base_dest_cmd,
    vectara_base_dest_cmd,
]

# Make sure there are not overlapping names
dest_cmd_names = [b().cmd_name for b in base_dest_cmd_fns]
dest_duplicates = [item for item, count in collections.Counter(dest_cmd_names).items() if count > 1]
if dest_duplicates:
    raise ValueError(
        "multiple base dest commands defined with the same names: {}".format(
            ", ".join(dest_duplicates),
        ),
    )

__all__ = [
    "base_src_cmd_fns",
    "base_dest_cmd_fns",
]



GOOGLE_DRIVE_EXPORT_TYPES = {
    "application/vnd.google-apps.document": "application/"
    "vnd.openxmlformats-officedocument.wordprocessingml.document",
    "application/vnd.google-apps.spreadsheet": "application/"
    "vnd.openxmlformats-officedocument.spreadsheetml.sheet",
    "application/vnd.google-apps.presentation": "application/"
    "vnd.openxmlformats-officedocument.presentationml.presentation",
    "application/vnd.google-apps.photo": "image/jpeg",
}



import typing as t
from dataclasses import MISSING, Field

from unstructured_ingest.enhanced_dataclass.core import _recursive_repr


class EnhancedField(Field):
    def __init__(self, *args, sensitive=False, overload_name: t.Optional[str] = None):
        super().__init__(*args)
        self.sensitive = sensitive
        self.overload_name = overload_name

    @_recursive_repr
    def __repr__(self):
        # Support for kw_only added in 3.10, to support as low as 3.8, need to dynamically map
        fields_array = [
            f"name={self.name!r}",
            f"type={self.type!r}",
            f"default={self.default!r}",
            f"default_factory={self.default_factory!r}",
            f"init={self.init!r}",
            f"repr={self.repr!r}",
            f"hash={self.hash!r}",
            f"compare={self.compare!r}",
            f"metadata={self.metadata!r}",
            f"sensitive={self.sensitive!r}",
            f"overload_name={self.overload_name!r}",
            f"_field_type={self._field_type}",
        ]
        if kw_only := getattr(self, "kw_only", None):
            fields_array.append(f"kw_only={kw_only!r}")
        return "Field({})".format(",".join(fields_array))


def enhanced_field(
    *,
    default=MISSING,
    default_factory=MISSING,
    init: bool = True,
    repr: bool = True,
    hash=None,
    compare: bool = True,
    metadata=None,
    kw_only=MISSING,
    sensitive: bool = False,
    overload_name: t.Optional[str] = None,
):
    if default is not MISSING and default_factory is not MISSING:
        raise ValueError("cannot specify both default and default_factory")
    args = [default, default_factory, init, repr, hash, compare, metadata]
    # Support for kw_only added in 3.10, to support as low as 3.8, need to dynamically map
    if "kw_only" in EnhancedField.__slots__:
        args.append(kw_only)
    return EnhancedField(*args, sensitive=sensitive, overload_name=overload_name)



import typing as t

import click

from unstructured_ingest.cli.cmds import base_dest_cmd_fns, base_src_cmd_fns

src: t.List[click.Group] = [v().get_src_cmd() for v in base_src_cmd_fns]

dest: t.List[click.Command] = [v().get_dest_cmd() for v in base_dest_cmd_fns]

__all__ = [
    "src",
    "dest",
]



from __future__ import annotations

from unstructured_ingest.v2.processes.connector_registry import (
    add_destination_entry,
)

from .duckdb import CONNECTOR_TYPE as DUCKDB_CONNECTOR_TYPE
from .duckdb import duckdb_destination_entry
from .motherduck import CONNECTOR_TYPE as MOTHERDUCK_CONNECTOR_TYPE
from .motherduck import motherduck_destination_entry

add_destination_entry(destination_type=DUCKDB_CONNECTOR_TYPE, entry=duckdb_destination_entry)
add_destination_entry(
    destination_type=MOTHERDUCK_CONNECTOR_TYPE, entry=motherduck_destination_entry
)



from __future__ import annotations

import hashlib
import logging
from pathlib import Path

from unstructured_ingest.interfaces import (
    ReadConfig,
)


def update_download_dir_remote_url(
    connector_name: str,
    read_config: ReadConfig,
    remote_url: str,
    logger: logging.Logger,
) -> str:
    hashed_dir_name = hashlib.sha256(remote_url.encode("utf-8"))
    return update_download_dir_hash(
        connector_name=connector_name,
        read_config=read_config,
        hashed_dir_name=hashed_dir_name,
        logger=logger,
    )


def update_download_dir_hash(
    connector_name: str,
    read_config: ReadConfig,
    hashed_dir_name: hashlib._Hash,
    logger: logging.Logger,
) -> str:
    if not read_config.download_dir:
        cache_path = Path.home() / ".cache" / "unstructured" / "ingest"
        if not cache_path.exists():
            cache_path.mkdir(parents=True, exist_ok=True)
        download_dir = cache_path / connector_name / hashed_dir_name.hexdigest()[:10]
        if read_config.preserve_downloads:
            logger.warning(
                f"Preserving downloaded files but download_dir is not specified,"
                f" using {download_dir}",
            )
        new_download_dir = str(download_dir)
        logger.debug(f"updating download directory to: {new_download_dir}")
    else:
        new_download_dir = read_config.download_dir
    return new_download_dir



import pytest
from pydantic import ValidationError

from unstructured_ingest.v2.processes.connectors.confluence import (
    ConfluenceAccessConfig,
    ConfluenceConnectionConfig,
)


def test_connection_config_multiple_auth():
    with pytest.raises(ValidationError):
        ConfluenceConnectionConfig(
            access_config=ConfluenceAccessConfig(
                password="api_token",
                token="access_token",
            ),
            username="user_email",
            url="url",
        )


def test_connection_config_no_auth():
    with pytest.raises(ValidationError):
        ConfluenceConnectionConfig(access_config=ConfluenceAccessConfig(), url="url")


def test_connection_config_basic_auth():
    ConfluenceConnectionConfig(
        access_config=ConfluenceAccessConfig(password="api_token"),
        url="url",
        username="user_email",
    )


def test_connection_config_pat_auth():
    ConfluenceConnectionConfig(
        access_config=ConfluenceAccessConfig(token="access_token"),
        url="url",
    )



from dataclasses import dataclass, field
from pathlib import Path
from typing import TYPE_CHECKING, Any, Generator, Optional
from uuid import NAMESPACE_DNS, uuid5

import pandas
from pydantic import BaseModel, Field, Secret, field_validator

from unstructured_ingest.utils.dep_check import requires_dependencies
from unstructured_ingest.v2.interfaces import (
    AccessConfig,
    ConnectionConfig,
    Downloader,
    DownloaderConfig,
    DownloadResponse,
    FileData,
    Indexer,
    IndexerConfig,
    SourceIdentifiers,
)
from unstructured_ingest.v2.processes.connector_registry import (
    SourceRegistryEntry,
)

if TYPE_CHECKING:
    from pyairtable import Api
    from pyairtable.api.types import RecordDict

CONNECTOR_TYPE = "airtable"


class AirtableTableMeta(BaseModel):
    """Metadata specifying a table id, a base id which the table is stored in,
    and an t.Optional view id in case particular rows and fields are to be ingested"""

    base_id: str
    table_id: str
    view_id: Optional[str] = None

    def get_id(self) -> str:
        id_s = f"{self.base_id}{self.table_id}"
        id_s = f"{id_s}{self.view_id}" if self.view_id else id_s
        return str(uuid5(NAMESPACE_DNS, id_s))


class AirtableAccessConfig(AccessConfig):
    personal_access_token: str = Field(
        description="Personal access token to authenticate into Airtable. Check: "
        "https://support.airtable.com/docs/creating-and-using-api-keys-and-access-tokens "
        "for more info"
    )


class AirtableConnectionConfig(ConnectionConfig):
    access_config: Secret[AirtableAccessConfig]

    @requires_dependencies(["pyairtable"], extras="airtable")
    def get_client(self) -> "Api":
        from pyairtable import Api

        access_config = self.access_config.get_secret_value()
        return Api(api_key=access_config.personal_access_token)


class AirtableIndexerConfig(IndexerConfig):
    list_of_paths: Optional[list[str]] = Field(
        default=None,
        description="""
        A list of paths that specify the locations to ingest data from within Airtable.

        If this argument is not set, the connector ingests all tables within each and every base.
        --list-of-paths: path1 path2 path3 ….
        path: base_id/table_id(optional)/view_id(optional)/

        To obtain (base, table, view) ids in bulk, check:
        https://airtable.com/developers/web/api/list-bases (base ids)
        https://airtable.com/developers/web/api/get-base-schema (table and view ids)
        https://pyairtable.readthedocs.io/en/latest/metadata.html (base, table and view ids)

        To obtain specific ids from Airtable UI, go to your workspace, and copy any
        relevant id from the URL structure:
        https://airtable.com/appAbcDeF1ghijKlm/tblABcdEfG1HIJkLm/viwABCDEfg6hijKLM
        appAbcDeF1ghijKlm -> base_id
        tblABcdEfG1HIJkLm -> table_id
        viwABCDEfg6hijKLM -> view_id

        You can also check: https://support.airtable.com/docs/finding-airtable-ids

        Here is an example for one --list-of-paths:
            base1/		→ gets the entirety of all tables inside base1
            base1/table1		→ gets all rows and columns within table1 in base1
            base1/table1/view1	→ gets the rows and columns that are
                                  visible in view1 for the table1 in base1

        Examples to invalid airtable_paths:
            table1          → has to mention base to be valid
            base1/view1     → has to mention table to be valid
                """,
    )

    @classmethod
    def validate_path(cls, path: str):
        components = path.split("/")
        if len(components) > 3:
            raise ValueError(
                f"Path must be of the format: base_id/table_id/view_id, "
                f"where table id and view id are optional. Got: {path}"
            )

    @field_validator("list_of_paths")
    @classmethod
    def validate_format(cls, v: list[str]) -> list[str]:
        for path in v:
            cls.validate_path(path=path)
        return v


@dataclass
class AirtableIndexer(Indexer):
    connector_type: str = CONNECTOR_TYPE
    connection_config: AirtableConnectionConfig
    index_config: AirtableIndexerConfig

    def get_all_table_meta(self) -> list[AirtableTableMeta]:
        client = self.connection_config.get_client()
        bases = client.bases()
        airtable_meta = []
        for base in bases:
            for table in base.schema().tables:
                airtable_meta.append(AirtableTableMeta(base_id=base.id, table_id=table.id))
        return airtable_meta

    def get_base_tables_meta(self, base_id: str) -> list[AirtableTableMeta]:
        client = self.connection_config.get_client()
        base = client.base(base_id=base_id)
        airtable_meta = []
        for table in base.tables():
            airtable_meta.append(AirtableTableMeta(base_id=base.id, table_id=table.id))
        return airtable_meta

    def get_meta_from_list(self) -> list[AirtableTableMeta]:
        airtable_meta = []
        for path in self.index_config.list_of_paths:
            components = path.split("/")
            if len(components) == 1:
                airtable_meta.extend(self.get_base_tables_meta(base_id=components[0]))
            elif len(components) == 2:
                airtable_meta.append(
                    AirtableTableMeta(base_id=components[0], table_id=components[1])
                )
            elif len(components) == 3:
                airtable_meta.append(
                    AirtableTableMeta(
                        base_id=components[0], table_id=components[1], view_id=components[2]
                    )
                )
            else:
                raise ValueError(
                    f"Path must be of the format: base_id/table_id/view_id, "
                    f"where table id and view id are optional. Got: {path}"
                )
        return airtable_meta

    def get_table_metas(self) -> list[AirtableTableMeta]:
        if not self.index_config.list_of_paths:
            return self.get_all_table_meta()
        return self.get_meta_from_list()

    def precheck(self) -> None:
        client = self.connection_config.get_client()
        client.request(method="HEAD", url=client.build_url("meta", "bases"))

    def run(self, **kwargs: Any) -> Generator[FileData, None, None]:
        table_metas = self.get_table_metas()
        for table_meta in table_metas:
            fullpath = (
                f"{table_meta.base_id}/{table_meta.table_id}/{table_meta.view_id}.csv"
                if table_meta.view_id
                else f"{table_meta.base_id}/{table_meta.table_id}.csv"
            )
            yield FileData(
                identifier=table_meta.get_id(),
                connector_type=CONNECTOR_TYPE,
                additional_metadata=table_meta.model_dump(),
                source_identifiers=SourceIdentifiers(
                    filename=str(Path(fullpath).name),
                    fullpath=fullpath,
                ),
            )


class AirtableDownloaderConfig(DownloaderConfig):
    pass


@dataclass
class AirtableDownloader(Downloader):
    connection_config: AirtableConnectionConfig
    download_config: AirtableDownloaderConfig = field(default_factory=AirtableDownloaderConfig)
    connector_type: str = CONNECTOR_TYPE

    def get_table_contents(self, table_meta: AirtableTableMeta) -> list["RecordDict"]:
        client = self.connection_config.get_client()
        table = client.table(base_id=table_meta.base_id, table_name=table_meta.table_id)
        table_fetch_kwargs = {"view": table_meta.view_id} if table_meta.view_id else {}
        rows = table.all(**table_fetch_kwargs)
        return rows

    def _table_row_to_dict(self, table_row: "RecordDict") -> dict:
        row_dict = {
            "id": table_row["id"],
            "created_time": table_row["createdTime"],
        }
        row_dict.update(table_row["fields"])
        return row_dict

    def run(self, file_data: FileData, **kwargs: Any) -> DownloadResponse:
        table_meta = AirtableTableMeta.model_validate(file_data.additional_metadata)
        table_contents = self.get_table_contents(table_meta=table_meta)
        df = pandas.DataFrame.from_dict(
            data=[self._table_row_to_dict(table_row=row) for row in table_contents]
        ).sort_index(axis=1)
        download_path = self.get_download_path(file_data=file_data)
        download_path.parent.mkdir(parents=True, exist_ok=True)
        df.to_csv(path_or_buf=download_path)
        return self.generate_download_response(file_data=file_data, download_path=download_path)


airtable_source_entry = SourceRegistryEntry(
    indexer=AirtableIndexer,
    indexer_config=AirtableIndexerConfig,
    downloader=AirtableDownloader,
    downloader_config=AirtableDownloaderConfig,
    connection_config=AirtableConnectionConfig,
)



import json
import os.path
from datetime import date, datetime
from gettext import gettext, ngettext
from gettext import gettext as _
from pathlib import Path
from typing import Any, Optional, Type, TypeVar, Union

import click
from pydantic import BaseModel, ConfigDict, Secret, TypeAdapter, ValidationError


def conform_click_options(options: dict):
    # Click sets all multiple fields as tuple, this needs to be updated to list
    for k, v in options.items():
        if isinstance(v, tuple):
            options[k] = list(v)


class Dict(click.ParamType):
    name = "dict"

    def convert(
        self,
        value: Any,
        param: Optional[click.Parameter] = None,
        ctx: Optional[click.Context] = None,
    ) -> Any:
        try:
            if isinstance(value, dict):
                return value
            if isinstance(value, Path) and value.is_file():
                with value.open() as f:
                    return json.load(f)
            if isinstance(value, str):
                return json.loads(value)
        except json.JSONDecodeError:
            self.fail(
                gettext(
                    "{value} is not a valid json value.",
                ).format(value=value),
                param,
                ctx,
            )


class FileOrJson(click.ParamType):
    name = "file-or-json"

    def __init__(self, allow_raw_str: bool = False):
        self.allow_raw_str = allow_raw_str

    def convert(
        self,
        value: Any,
        param: Optional[click.Parameter] = None,
        ctx: Optional[click.Context] = None,
    ) -> Any:
        # check if valid file
        full_path = os.path.abspath(os.path.expanduser(value))
        if os.path.isfile(full_path):
            return str(Path(full_path).resolve())
        if isinstance(value, str):
            try:
                return json.loads(value)
            except json.JSONDecodeError:
                if self.allow_raw_str:
                    return value
        self.fail(
            gettext(
                "{value} is neither a valid json string nor an existing filepath.",
            ).format(value=value),
            param,
            ctx,
        )


class DelimitedString(click.ParamType):
    name = "delimited-string"

    def __init__(self, delimiter: str = ",", choices: Optional[list[str]] = None):
        self.choices = choices if choices else []
        self.delimiter = delimiter

    def convert(
        self,
        value: Any,
        param: Optional[click.Parameter] = None,
        ctx: Optional[click.Context] = None,
    ) -> Any:
        # In case a list is provided as the default, will not break
        if isinstance(value, list):
            split = [str(v).strip() for v in value]
        else:
            split = [v.strip() for v in value.split(self.delimiter)]
        if not self.choices:
            return split
        choices_str = ", ".join(map(repr, self.choices))
        for s in split:
            if s not in self.choices:
                self.fail(
                    ngettext(
                        "{value!r} is not {choice}.",
                        "{value!r} is not one of {choices}.",
                        len(self.choices),
                    ).format(value=s, choice=choices_str, choices=choices_str),
                    param,
                    ctx,
                )
        return split


class PydanticDateTime(click.ParamType):
    name = "datetime"

    def convert(
        self,
        value: Any,
        param: Optional[click.Parameter] = None,
        ctx: Optional[click.Context] = None,
    ) -> Any:
        try:
            return TypeAdapter(datetime).validate_strings(value)
        except ValidationError:
            self.fail(f"{value} is not a valid datetime", param, ctx)


class PydanticDate(click.ParamType):
    name = "date"

    def convert(
        self,
        value: Any,
        param: Optional[click.Parameter] = None,
        ctx: Optional[click.Context] = None,
    ) -> Any:
        try:
            return TypeAdapter(date).validate_strings(value)
        except ValidationError:
            self.fail(f"{value} is not a valid date", param, ctx)


BaseModelT = TypeVar("BaseModelT", bound=BaseModel)


def unwrap_optional(val: Any) -> tuple[Any, bool]:
    if (
        hasattr(val, "__origin__")
        and hasattr(val, "__args__")
        and val.__origin__ is Union
        and len(val.__args__) == 2
        and type(None) in val.__args__
    ):
        args = val.__args__
        args = [a for a in args if a is not None]
        return args[0], True
    return val, False


def extract_config(flat_data: dict, config: Type[BaseModelT]) -> BaseModelT:
    fields = config.model_fields
    config.model_config = ConfigDict(extra="ignore")
    field_names = [v.alias or k for k, v in fields.items()]
    data = {k: v for k, v in flat_data.items() if k in field_names and v is not None}
    if access_config := fields.get("access_config"):
        access_config_type = access_config.annotation
        access_config_type, is_optional = unwrap_optional(access_config_type)
        # Check if raw type is wrapped by a secret
        if (
            hasattr(access_config_type, "__origin__")
            and hasattr(access_config_type, "__args__")
            and access_config_type.__origin__ is Secret
        ):
            ac_subtypes = access_config_type.__args__
            ac_fields = ac_subtypes[0].model_fields
        elif issubclass(access_config_type, BaseModel):
            ac_fields = access_config_type.model_fields
        else:
            raise TypeError(f"Unrecognized access_config type: {access_config_type}")
        ac_field_names = [v.alias or k for k, v in ac_fields.items()]
        access_config_data = {
            k: v for k, v in flat_data.items() if k in ac_field_names and v is not None
        }
        if not access_config_data and is_optional:
            data["access_config"] = None
        else:
            data["access_config"] = access_config_data
    return config.model_validate(obj=data)


class Group(click.Group):
    def parse_args(self, ctx, args):
        """
        This allows for subcommands to be called with the --help flag without breaking
        if parent command is missing any of its required parameters
        """
        try:
            return super().parse_args(ctx, args)
        except click.MissingParameter:
            if "--help" not in args:
                raise
            # remove the required params so that help can display
            for param in self.params:
                param.required = False
            return super().parse_args(ctx, args)

    def format_commands(self, ctx: click.Context, formatter: click.HelpFormatter) -> None:
        """
        Copy of the original click.Group format_commands() method but replacing
        'Commands' -> 'Destinations'
        """
        commands = []
        for subcommand in self.list_commands(ctx):
            cmd = self.get_command(ctx, subcommand)
            # What is this, the tool lied about a command.  Ignore it
            if cmd is None:
                continue
            if cmd.hidden:
                continue

            commands.append((subcommand, cmd))

        # allow for 3 times the default spacing
        if len(commands):
            if formatter.width:
                limit = formatter.width - 6 - max(len(cmd[0]) for cmd in commands)
            else:
                limit = -6 - max(len(cmd[0]) for cmd in commands)

            rows = []
            for subcommand, cmd in commands:
                help = cmd.get_short_help_str(limit)
                rows.append((subcommand, help))

            if rows:
                with formatter.section(_("Destinations")):
                    formatter.write_dl(rows)
